import requests
import json
from citrination_client.util.quote_finder import quote
from citrination_client.base.response_handling import raise_on_response, check_general_success, check_for_rate_limiting, get_response_json
from citrination_client.base.errors import *

from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


DEFAULT_FAILURE_MESSAGE = "There was an error communicating with Citrination"

class BaseClient(object):
    """
    Base class that holds the universal constructor, utilities, etc
    """

    def __init__(self, api_key, webserver_host, api_members=[], suppress_warnings=False):
        """
        Constructor.

        :param api_key: Authentication token for the Citrination site
        :type api_key: str
        :param host: The base URL of the citrination site, e.g. https://citrination.com
        :type host: str
        :param api_members: The names of the member methods for this client
        :type api_members: str[]
        :param suppress_warnings: A flag indicating whether or not warning
            messages to stdout should be printed
        :type suppress_warnings: bool
        """
        if api_key == None or len(api_key) == 0:
            raise CitrinationClientError("API key must be present to instantiate the client")

        self.headers = {
            'X-API-Key': quote(api_key),
            'Content-Type': 'application/json',
            'X-Citrination-API-Version': '1.0.0'
        }
        self.suppress_warnings = suppress_warnings
        self.api_url = webserver_host + '/api'
        self.api_members = api_members

    # ==== Private Utilities ===

    def _handle_response(self, response, failure_message=DEFAULT_FAILURE_MESSAGE):
        raise_on_response(response)
        check_general_success(response, failure_message)
        return response

    def _get_success_json(self, response):
        return get_response_json(response)

    def _get_qualified_route(self, route):
        """
        Get a fully qualified api route.
        :param route: the route (e.g., /model)
        :return: the fully qualified route (e.g., https://citrination.com/model)
        """
        return "{}/{}".format(self.api_url, route)

    def _warn(self, message):
        if not self.suppress_warnings:
            print("Citrination Client Warning - {}".format(message))

    def _get_headers(self, headers=None):
        if headers:
            return headers
        else:
            return self.headers

    def _get(self, route, headers=None, failure_message=None):
        """
        Execute a post request and return the result
        :param headers:
        :return:
        """
        headers = self._get_headers(headers)
        response_lambda = (lambda: requests.get(self._get_qualified_route(route), headers=headers, verify=False))
        response = check_for_rate_limiting(response_lambda(), response_lambda)
        return self._handle_response(response, failure_message)

    def _post_json(self, route, data, headers=None, failure_message=None):
        return self._post(route, json.dumps(data), headers)

    def _post(self, route, data, headers=None, failure_message=None):
        """
        Execute a post request and return the result
        :param data:
        :param headers:
        :return:
        """
        headers = self._get_headers(headers)
        response_lambda = (lambda: requests.post(self._get_qualified_route(route), headers=headers, data=data, verify=False))
        response = check_for_rate_limiting(response_lambda(), response_lambda)
        return self._handle_response(response, failure_message)

    def _put_json(self, route, data, headers=None, failure_message=None):
        return self._put(route, json.dumps(data), headers)

    def _put(self, route, data, headers=None, failure_message=None):
        """
        Execute a put request and return the result
        :param data:
        :param headers:
        :return:
        """
        headers = self._get_headers(headers)
        response_lambda = (lambda: requests.put(self._get_qualified_route(route), headers=headers, data=data, verify=False))
        response = check_for_rate_limiting(response_lambda(), response_lambda)
        return self._handle_response(response, failure_message)

    def _delete(self, route, headers=None, failure_message=None):
        """
        Execute a delete request and return the result
        :param headers:
        :return:
        """
        headers = self._get_headers(headers)
        response_lambda = (lambda: requests.delete(self._get_qualified_route(route), headers=headers, verify=False))
        response = check_for_rate_limiting(response_lambda(), response_lambda)
        return self._handle_response(response, failure_message)

    def __repr__(self):
        return "{}".format(self.api_members)


def data_analysis(view_id):
  return 'data_views/{}/data_analysis'.format(view_id)

def data_view_predict(view_id):
    return 'data_views/{}/predict'.format(view_id)

def custom_model_predict(model_path):
    return 'ml_templates/{}/predict'.format(model_path)

def submit_data_view_design(data_view_id):
    return "data_views/{}/experimental_design".format(data_view_id)

def get_data_view_design_status(data_view_id, run_uuid):
    return "data_views/{}/experimental_design/{}/status".format(data_view_id, run_uuid)

def get_data_view_design_results(data_view_id, run_uuid):
    return "data_views/{}/experimental_design/{}/results".format(data_view_id, run_uuid)

def kill_data_view_design_run(data_view_id, run_uuid):
    return "data_views/{}/experimental_design/{}".format(data_view_id, run_uuid)

def get_data_view(data_view_id):
    return "data_views/{}".format(data_view_id)



from citrination_client.models.columns.base import BaseColumn
from citrination_client.base.errors import CitrinationClientError

class RealColumn(BaseColumn):
    """
    A real column configuration for a data view. Parameterized
    with the basic column options and an upper and lower bound on the
    acceptable values for the column
    """

    TYPE = "Real"

    def __init__(self, name, role, group_by_key=False, units=None, lower_bound=None, upper_bound=None):
        """
        Constructor.

        :param name: The name of the column
        :type name: str
        :param role: The role the column will play in machine learning:
                       "Input"
                       "Output"
                       "Latent Variable"
                       "Ignore"
        :type role: str
        :param group_by_key: Whether or not this column should be used for
            grouping during cross validation
        :type group_by_key: bool
        :param units: Optionally, the units for the column
        :type units: str
        :param lower_bound: The lower bound for valid values for this column
        :type lower_bound: float
        :param upper_bound: The upper bound for valid values for this column
        :type upper_bound: float
        """
        super(RealColumn, self).__init__(name=name,
                                         role=role,
                                         group_by_key=group_by_key,
                                         units=units)
        # Default bounds to None to enable validation via the setter methods
        self._lower_bound = None
        self._upper_bound = None

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def build_options(self):
        # Ensure that infinity values are serialized
        # into "Infinity" or "-Infinity", as the Citrination
        # server expects
        if self.upper_bound == float("inf"):
            upper_bound = "Infinity"
        else:
            upper_bound = self.upper_bound

        if self.lower_bound == float("-inf"):
            lower_bound = "-Infinity"
        else:
            lower_bound = self.lower_bound

        return {
            "lower_bound": lower_bound,
            "upper_bound": upper_bound
        }

    @property
    def lower_bound(self):
        return self._lower_bound

    @lower_bound.setter
    def lower_bound(self, value):
        self._lower_bound = self._cast_and_validate_float(value, "lower_bound")
        self._validate_bounds()

    @lower_bound.deleter
    def lower_bound(self):
        self._lower_bound = None

    @property
    def upper_bound(self):
        return self._upper_bound

    @upper_bound.setter
    def upper_bound(self, value):
        self._upper_bound = self._cast_and_validate_float(value, "upper_bound")
        self._validate_bounds()

    @upper_bound.deleter
    def upper_bound(self):
        self._upper_bound = None

    def _both_bounds_present(self):
        return self.lower_bound is not None and self.upper_bound is not None

    def _validate_bounds(self):
        if self._both_bounds_present() and self._lower_bound > self._upper_bound:
            raise CitrinationClientError("When constructing a RealColumn, lower_bound must be less than upper_bound")

    def _cast_and_validate_float(self, value, attr_name):
        try:
            return float(value)
        except ValueError:
            raise CitrinationClientError("For a RealColumn, {} must be castable as a float".format(attr_name))



from citrination_client.search.core.query.base_returning_query import BaseReturningQuery


class FileReturningQuery(BaseReturningQuery):
    """
    Query used to return information about files.
    """

    def __init__(self, query=None, from_index=None, size=None, random_results=None, random_seed=None,
                 score_relevance=None, return_max_score=None, timeout=None, max_content_highlights=None,
                 highlight_pre_tag=None, highlight_post_tag=None, **kwargs):
        """
        Constructor.

        :param query: One or more :class:`DataQuery` objects with the queries to run.
        :param from_index: Index of the first hit that should be returned.
        :param size: Total number of hits the should be returned.
        :param random_results: Whether to return a random set of records.
        :param random_seed: The random seed to use.
        :param score_relevance: Whether to use relevance scoring.
        :param return_max_score: Whether to return the maximum score.
        :param timeout: The number of milliseconds to wait for the query to execute.
        :param max_content_highlights: The maximum number of highlighted results to return.
        :param highlight_pre_tag: The tag to use at the beginning of a highlight.
        :param highlight_post_tag: The tag to use at the end of a highlight.
        """
        super(FileReturningQuery, self).__init__(
            query=query, from_index=from_index, size=size, random_results=random_results, random_seed=random_seed,
            score_relevance=score_relevance, return_max_score=return_max_score, timeout=timeout, **kwargs)
        self._max_content_highlights = None
        self.max_content_highlights = max_content_highlights
        self._highlight_pre_tag = None
        self.highlight_pre_tag = highlight_pre_tag
        self._highlight_post_tag = None
        self.highlight_post_tag = highlight_post_tag

    @property
    def max_content_highlights(self):
        return self._max_content_highlights

    @max_content_highlights.setter
    def max_content_highlights(self, max_content_highlights):
        self._max_content_highlights = max_content_highlights

    @max_content_highlights.deleter
    def max_content_highlights(self):
        self._max_content_highlights = None

    @property
    def highlight_pre_tag(self):
        return self._highlight_pre_tag

    @highlight_pre_tag.setter
    def highlight_pre_tag(self, highlight_pre_tag):
        self._highlight_pre_tag = highlight_pre_tag

    @highlight_pre_tag.deleter
    def highlight_pre_tag(self):
        self._highlight_pre_tag = None

    @property
    def highlight_post_tag(self):
        return self._highlight_post_tag

    @highlight_post_tag.setter
    def highlight_post_tag(self, highlight_post_tag):
        self._highlight_post_tag = highlight_post_tag

    @highlight_post_tag.deleter
    def highlight_post_tag(self):
        self._highlight_post_tag = None



class Projection(object):
    """
    A projection to be included in the TSNE analysis.
    """

    def __init__(self, xs, ys, responses, tags, uids):
        """
        Constructor.

        :param xs: A list of x values of the projection.
        :type xs: list of floats
        :param ys: A list of y values of the projection.
        :type ys: list of floats
        :param responses: A list of z values of the projection.
        :type responses: list of floats
        :param tags: A list of tags for the projected points
        :type tags: list of strings
        :param uids: A list of record UIDs for the projected points
        :type uids: list of strings
        """
        self._xs = xs
        self._ys = ys
        self._responses = responses
        self._tags = tags
        self._uids = uids

    @property
    def xs(self):
        return self._xs

    @property
    def ys(self):
        return self._ys

    @property
    def responses(self):
        return self._responses

    @property
    def tags(self):
        return self._tags

    @property
    def uids(self):
        return self._uids


from citrination_client.models.columns.categorical import CategoricalColumn
from citrination_client.models.columns.real import RealColumn
from citrination_client.models.columns.vector import VectorColumn
from citrination_client.models.columns.alloy_composition import AlloyCompositionColumn
from citrination_client.models.columns.inorganic_chemical_formula import InorganicChemicalFormulaColumn
from citrination_client.models.columns.organic_chemical_formula import OrganicChemicalFormulaColumn

class DescriptorConverter(object):

    @classmethod
    def convert(self, col_name, descriptor, role, units=None):
        column_type = descriptor['category']
        if column_type == CategoricalColumn.TYPE:
            categories = [str(v) for v in descriptor["descriptorValues"]]

            return CategoricalColumn(
                    name=col_name,
                    role=role,
                    units=units,
                    categories=categories
                )
        elif column_type == RealColumn.TYPE:
            lower_bound = descriptor["lowerBound"]
            upper_bound = descriptor["upperBound"]

            return RealColumn(
                    name=col_name,
                    role=role,
                    units=units,
                    lower_bound=lower_bound,
                    upper_bound=upper_bound
                )
        elif column_type == AlloyCompositionColumn.TYPE:
            balance_element = descriptor['balanceElement']
            basis           = descriptor['basis']

            return AlloyCompositionColumn(
                    name=col_name,
                    role=role,
                    units=units,
                    balance_element=balance_element,
                    basis=basis
                )
        elif column_type == InorganicChemicalFormulaColumn.TYPE:
            return InorganicChemicalFormulaColumn(
                    name=col_name,
                    role=role,
                    units=units
                )
        elif column_type == OrganicChemicalFormulaColumn.TYPE:
            return OrganicChemicalFormulaColumn(
                    name=col_name,
                    role=role,
                    units=units
                )
        elif column_type == VectorColumn.TYPE:
            length = descriptor["length"]

            return VectorColumn(
                    name=col_name,
                    role=role,
                    units=units,
                    length=length
                )



class UploadResult(object):
    """
    The result of an attempted upload. Keeps track of the failures
    and successes if multiple files were uploaded (for instance,
    if a directory was uploaded).
    """

    def __init__(self):
        """
        Constructor.
        """
        self._failures = []
        self._successes = []

    @property
    def failures(self):
        return self._failures

    @property
    def successes(self):
        return self._successes

    def successful(self):
        """
        Indicates whether or not the entire upload was successful.

        :return: Whether or not the upload was successful
        :rtype: bool
        """
        return len(self._failures) == 0

    def add_failure(self, filepath, reason):
        """
        Registers a file as a failure to upload.

        :param filepath: The path to the file which was to be uploaded.
        :type filepath: str
        :param reason: The reason the file failed to upload
        :type reason: str
        """
        self._failures.append({
                "path": filepath,
                "reason": reason
            })

    def add_success(self, filepath):
        """
        Registers a file as successfully uploaded.

        :param filepath: The path to the successfully uploaded file.
        :type filepath: str
        """
        self._successes.append({
                "path": filepath
            })


class DatasetFile(object):
    """
    Class representation of a file in a dataset on Citrination.
    """

    def __init__(self, path, url=None):
        """
        Constructor.

        :param path: The files path
        :type path: str
        :param url: If present, a download URL for the file
        :type url: str
        """
        self._path = path
        self._url = url

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value

    @path.deleter
    def path(self):
        self._path = None

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @url.deleter
    def url(self):
        self._url = None


from citrination_client.search.core.result.base_search_result import BaseSearchResult
from citrination_client.search.file.result.file_search_hit import FileSearchHit


class FileSearchResult(BaseSearchResult):
    """
    Class to store the results of a file query.
    """

    def __init__(self, took=None, total_num_hits=None, max_score=None, hits=None, **kwargs):
        """
        Constructor.

        :param took: Number of milliseconds that the query took to execute.
        :param total_num_hits: Total number of hits.
        :param max_score: The maximum score.
        :param hits: List of :class:`FileSearchHit` objects.
        """
        super(FileSearchResult, self).__init__(
            took=took, total_num_hits=total_num_hits, max_score=max_score,
            hits=self._get_object(FileSearchHit, hits), **kwargs)



from citrination_client.models.columns import *
from citrination_client.base import CitrinationClientError

def _flatten_column_dict(response_dict):
    flat_dict = {}
    for k in response_dict:
        if k == "options" or k == "type":
            continue
        flat_dict[k] = response_dict[k]

    if response_dict["options"]:
        for option in response_dict["options"]:
            flat_dict[option] = response_dict["options"][option]

    return flat_dict

def _get_column_class_from_type(type):
    if type == RealColumn.TYPE:
        return RealColumn
    elif type == CategoricalColumn.TYPE:
        return CategoricalColumn
    elif type == AlloyCompositionColumn.TYPE:
        return AlloyCompositionColumn
    elif type == InorganicChemicalFormulaColumn.TYPE:
        return InorganicChemicalFormulaColumn
    elif type == VectorColumn.TYPE:
        return VectorColumn
    else:
        raise CitrinationClientError("Unknown column type: {}".format(type))

class ColumnFactory(object):

    @staticmethod
    def from_dict(response_dict):
        column_class = _get_column_class_from_type(response_dict["type"])
        processed_dict = _flatten_column_dict(response_dict)
        return column_class(**processed_dict)



from citrination_client.models.columns.base import BaseColumn

class OrganicChemicalFormulaColumn(BaseColumn):
    """
    An organic chemical formula column configuration for a data view.
    Parameterized with the basic column options.
    """

    TYPE = "Organic"



from citrination_client.base.errors import *
from time import sleep

def check_for_rate_limiting(response, response_lambda, timeout=1, attempts=0):
    """
    Takes an initial response, and a way to repeat the request that produced it and retries the request with an increasing sleep period between requests if rate limiting resposne codes are encountered.

    If more than 3 attempts are made, a RateLimitingException is raised

    :param response: A response from Citrination
    :type response: requests.Response
    :param response_lambda: a callable that runs the request that returned the
        response
    :type response_lambda: function
    :param timeout: the time to wait before retrying
    :type timeout: int
    :param attempts: the number of the retry being executed
    :type attempts: int
    """
    if attempts >= 3:
        raise RateLimitingException()
    if response.status_code == 429:
        sleep(timeout)
        new_timeout = timeout + 1
        new_attempts = attempts + 1
        return check_for_rate_limiting(response_lambda(timeout, attempts), response_lambda, timeout=new_timeout, attempts=new_attempts)
    return response

def check_general_success(response, failure_message):
    if response.status_code >= 400:
        raise CitrinationClientError(
            "{} - Citrination returned {}".format(failure_message, response.status_code)
        )
    return response

def get_response_json(response):
    return response.json()

def raise_on_response(response):
    _check_response_for_version_mismatch(response)
    _check_response_for_feature_availability(response)
    _check_response_for_authorization(response)
    _check_response_for_timeout(response)
    _check_response_for_server_error(response)
    return response

def _check_response_for_authorization(response):
    if response.status_code == 401:
        raise UnauthorizedAccessException()

def _check_response_for_feature_availability(response):
    if response.status_code == 403:
        raise FeatureUnavailableException()

def _check_response_for_timeout(response):
    if response.status_code == 524:
        raise RequestTimeoutException()

def _check_response_for_server_error(response):
    if response.status_code >= 500:
        raise CitrinationServerErrorException(
                "Citrination returned an error code - code: {}, body: {}".format(response.status_code, response.content)
            )

def _check_response_for_version_mismatch(response):
    try:
        if response.status_code == 400:
            error_type = response.json()["error_type"]
            if error_type == "Version Mismatch":
                raise APIVersionMismatchException()
        return response
    except (ValueError, KeyError):
        return response


class BaseConstraint(object):

    def to_dict(self):
        """
        Returns a dictionary representing the constraint.
        Assists in JSON serialization.

        :return: A dictionary with the name and type of the constraint
            along with any type-specific metadata the constraint may
            require
        """
        return {
            "name": self._name,
            "type": self._type,
            "options": self.options()
        }



from pypif.util.case import to_camel_case
from pypif.util.case import keys_to_snake_case

import json

class QueryEncoder(json.JSONEncoder):
    """
    Class used to convert a query to json.
    """

    def default(self, obj):
        """
        Convert an object to a form ready to dump to json.

        :param obj: Object being serialized. The type of this object must be one of the following: None; a single object derived from the Pio class; or a list of objects, each derived from the Pio class.
        :return: List of dictionaries, each representing a physical information object, ready to be serialized.
        """
        if obj is None:
            return []
        elif isinstance(obj, list):
            return [i.as_dictionary() for i in obj]
        elif isinstance(obj, dict):
            return self._keys_to_camel_case(obj)
        else:
            return obj.as_dictionary()

    def _keys_to_camel_case(self, obj):
        """
        Make a copy of a dictionary with all keys converted to camel case. This is just calls to_camel_case on each of the keys in the dictionary and returns a new dictionary.

        :param obj: Dictionary to convert keys to camel case.
        :return: Dictionary with the input values and all keys in camel case
        """
        return dict((to_camel_case(key), value) for (key, value) in obj.items())



from citrination_client.base.errors import CitrinationClientError

class BaseColumn(object):
    """
    Base class for column configuration in a data view. Subclasses
    each represent one type of column.
    """

    def __init__(self, name, role, group_by_key=False, units=None):
        """
        Constructor.

        :param name: The name of the column
        :type name: str
        :param role: The role the column will play in machine learning:
                       "Input"
                       "Output"
                       "Latent Variable"
                       "Ignore"
        :type role: str
        :param group_by_key: Whether or not this column should be used for
            grouping during cross validation
        :type group_by_key: bool
        :param units: Optionally, the units for the column
        :type units: str
        """
        self._name = name
        self._units = units
        self.role = role
        self._type = self.__class__.TYPE
        self._group_by_key = group_by_key

    def to_dict(self):
        """
        Converts the column to a dictionary representation accepted
        by the Citrination server.

        :return: Dictionary with basic options, plus any column type specific
            options held under the "options" key
        :rtype: dict
        """
        return {
            "type": self.type,
            "name": self.name,
            "group_by_key": self.group_by_key,
            "role": self.role,
            "units": self.units,
            "options": self.build_options()
        }

    def build_options(self):
        """
        Default value for optional column configuration. Only child
        classes will have non-None values for this field. Value
        depends on child class implementation.

        Note: Some child classes do not require any extra options
        (e.g. OrganicChemicalFormula), in that case, this implementation
        will be invoked and no options will be present in the dictionary.

        :return: Options dictionary, or None if not implemented in child
        :rtype: dict or None
        """
        return None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @name.deleter
    def name(self):
        self._name = None

    @property
    def role(self):
        return self._role

    @role.setter
    def role(self, value):
        self._validate_role(value)
        self._role = value

    @role.deleter
    def role(self):
        self._role = None

    @property
    def group_by_key(self):
        return self._group_by_key

    @group_by_key.setter
    def group_by_key(self, value):
        self._group_by_key = value

    @group_by_key.deleter
    def group_by_key(self):
        self._group_by_key = None

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        self._units = value

    @units.deleter
    def units(self):
        self._units = None

    @property
    def type(self):
        """
        Returns the type of the column. This value is
        read only because the valid value is defined by the
        column type and is set internally.
        """
        return self._type

    def _validate_role(self, role):
        """
        Validates that the role parameter has a valid value.

        :param role: Name of the column's role
        :type role: str
        """
        valid_roles = ["Input",
                       "Output",
                       "Latent Variable",
                       "Ignore",
                       "input",
                       "output",
                       "latentVariable",
                       "ignore"]
        if role not in valid_roles:
            raise CitrinationClientError("Column role must be one of: {}".format(valid_roles))



from citrination_client.models.columns.base import BaseColumn
from citrination_client.models.columns.categorical import CategoricalColumn
from citrination_client.models.columns.real import RealColumn
from citrination_client.models.columns.vector import VectorColumn
from citrination_client.models.columns.organic_chemical_formula import OrganicChemicalFormulaColumn
from citrination_client.models.columns.inorganic_chemical_formula import InorganicChemicalFormulaColumn
from citrination_client.models.columns.alloy_composition import AlloyCompositionColumn
from citrination_client.models.columns.descriptor_converter import DescriptorConverter



class ProcessStatus(object):
    """
    The status of an in progress process executing on Citrination.
    """

    def __init__(self, result, progress, status, messages=None):
        """
        Constructor.

        :param result: The result of the process
        :type result: any
        :param progress: The progress of the process as as percentage
        :type progress: int
        :param status: The status string for the process
        :type status: str
        :param messages: A list of messages representing the steps the process
            has already progressed through
        :type messages: list of str
        """
        self._status = status
        self._result = result
        self._progress = progress
        self._messages = messages

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @uuid.deleter
    def uuid(self):
        self._uuid = None

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, value):
        self._result = value

    @result.deleter
    def result(self):
        self._result = None

    @property
    def progress(self):
        return self._progress

    @progress.setter
    def progress(self, value):
        self._progress = value

    @progress.deleter
    def progress(self):
        self._progress = None

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

    @status.deleter
    def status(self):
        self._status = None

    @property
    def messages(self):
        return self._messages

    @messages.setter
    def messages(self, value):
        self._messages = value

    @messages.deleter
    def messages(self):
        self._messages = None

    def finished(self):
        return self.status == "Finished"

    def killed(self):
        return self.status == "Killed"

    def accepted(self):
        return self.status == "Accepted"


from citrination_client.util.quote_finder import quote

def update_file(file_id):
    return 'data_sets/update_file/{}'.format(file_id)

def list_files(dataset_id):
    return 'datasets/{}/list_filepaths'.format(dataset_id)

def matched_files(dataset_id):
    return 'datasets/{}/download_files'.format(dataset_id)

def pif_dataset_uid(dataset_id, pif_uid):
    return 'datasets/{}/pif/{}'.format(dataset_id, pif_uid)

def pif_dataset_version_uid(dataset_id, version, pif_uid):
    return 'datasets/{}/version/{}/pif/{}'.format(dataset_id, version, pif_uid)

def create_dataset():
    return 'data_sets/create_dataset'

def create_dataset_version(dataset_id):
    return 'data_sets/{}/create_dataset_version'.format(dataset_id)

def update_dataset(dataset_id):
    return 'data_sets/{}/update'.format(dataset_id)

def upload_to_dataset(dataset_id):
    return 'data_sets/{}/upload'.format(dataset_id)

def file_dataset_path(dataset_id, file_path):
    return 'datasets/{}/file/{}'.format(dataset_id, quote(file_path))

def file_dataset_version_path(dataset_id, version, file_path):
    return 'datasets/{}/version/{}/files/{}'.format(dataset_id, version, quote(file_path))


class DesignRun(object):
    """
    An in progress design run. Contains a UUID which is the identifier
    for the run.
    """

    def __init__(self, uuid):
        """
        Constructor.

        :param uuid: The UUID of an in progress design run.
        :type uuid: str
        """
        self._uuid = uuid

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @uuid.deleter
    def uuid(self):
        self._uuid = None


from citrination_client.search.core.query.base_returning_query import BaseReturningQuery


class DatasetReturningQuery(BaseReturningQuery):
    """
    Query used to return information about datasets. Citrination does not support pagination past the 50,000th result. Please chose values for from_index and size that do not exceed this limit.
    """

    def __init__(self, query=None, from_index=None, size=None, random_results=None, random_seed=None,
                 score_relevance=None, return_max_score=None, timeout=None, count_pifs=None, **kwargs):
        """
        Constructor.

        :param query: One or more :class:`DataQuery` objects with the queries to run.
        :param from_index: Index of the first hit that should be returned.
        :param size: Total number of hits the should be returned.
        :param random_results: Whether to return a random set of records.
        :param random_seed: The random seed to use.
        :param score_relevance: Whether to use relevance scoring.
        :param return_max_score: Whether to return the maximum score.
        :param timeout: The number of milliseconds to wait for the query to execute.
        :param count_pifs: Whether to return counts of PIFs for each dataset.
        """
        super(DatasetReturningQuery, self).__init__(
            query=query, from_index=from_index, size=size, random_results=random_results, random_seed=random_seed,
            score_relevance=score_relevance, return_max_score=return_max_score, timeout=timeout, **kwargs)
        self._count_pifs = None
        self.count_pifs = count_pifs

    @property
    def count_pifs(self):
        return self._count_pifs

    @count_pifs.setter
    def count_pifs(self, count_pifs):
        self._count_pifs = count_pifs

    @count_pifs.deleter
    def count_pifs(self):
        self._count_pifs = None



class PredictedValue(object):
    """
    The value/loss output from a prediction.
    """

    def __init__(self, key, value, loss=None):
        """
        Constructor.

        :param key: The descriptor key for the prediction
        :type key: str
        :param value: The predicted value
        :type value: str or float
        :param loss: The loss for the prediction
        :type loss: float
        """
        self._key = key
        self._value = value
        self._loss = loss

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        self._key = value

    @key.deleter
    def key(self):
        self._key = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @value.deleter
    def value(self):
        self._value = None

    @property
    def loss(self):
        return self._loss

    @loss.setter
    def loss(self, value):
        self._loss = value

    @loss.deleter
    def loss(self):
        self._loss = None


from setuptools import setup, find_packages

setup(name='citrination-client',
      version='4.0.1',
      url='http://github.com/CitrineInformatics/python-citrination-client',
      description='Python client for accessing the Citrination api',
      packages=find_packages(exclude=('docs')),
      install_requires=[
          'requests<3',
          'pypif',
          'six<2',
          'pyyaml'
      ],
      extras_require={
        "dev": [
          'sphinx_rtd_theme',
          'sphinx',
        ],
        "test": [
          'requests_mock',
          'pytest',
        ]
      })



class DataView(object):

    def __init__(self, view_id, name, description, datasets=[], column_names=[], columns=[]):
        """
        Constructor.

        :param view_id: The ID of the data view
        :type view_id: str
        :param name: The name of the data view
        :type name: str
        :param description: The description of the data view
        :type description: str
        :param datasets: The datasets used in the view
        :type datasets: list of :class: Dataset
        :param column_names: The column names in the view
        :type column_names: list of str
        """
        self._id = view_id
        self._name = name
        self._description = description
        self._datasets = datasets
        self._column_names = column_names
        self._columns = columns

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @id.deleter
    def id(self):
        self._id = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @name.deleter
    def name(self):
        self._name = None

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @description.deleter
    def description(self):
        self._description = None

    @property
    def datasets(self):
        return self._datasets

    @datasets.setter
    def datasets(self, value):
        self._datasets = value

    @datasets.deleter
    def datasets(self):
        self._datasets = None

    @property
    def column_names(self):
        return self._column_names

    @column_names.setter
    def column_names(self, value):
        self._column_names = value

    @column_names.deleter
    def column_names(self):
        self._column_names = None

    @property
    def columns(self):
        return self._columns

    @columns.setter
    def columns(self, value):
        self._columns = value

    @columns.deleter
    def columns(self):
        self._columns = None



from citrination_client.models import ModelsClient
from citrination_client.search import SearchClient
from citrination_client.data import DataClient
from citrination_client.util.credentials import get_preferred_credentials

"""
Generates a lambda method in a closure such that the
client and method_name are constant (avoides reassignment
in the assignment loop in the client class)
"""
def _generate_lambda_proxy_method(client, method_name):
    subclient_m = getattr(client, method_name)
    return lambda *args, **kw: subclient_m(*args, **kw)

class CitrinationClient(object):
    """
    The top level of the client hierarchy. Instantiating this class handles
    authentication information (api_key and site) and provides access to instances of each of the sub-clients, for more specific actions.

    Instantiation requires authentication information, but that can be provided
    via direct parameterization, environment variables, or a .citrination credentials file. See the tutorial on client Initialization for more information.
    """

    def __init__(self, api_key=None, site=None, suppress_warnings=False):
        """
        Constructor.

        :param api_key: Your API key for Citrination
        :type api_key: str
        :param site: The domain name of your Citrination deployment
            (the default is https://citrination.com)
        :type site: str
        :param suppress_warnings: A flag allowing you to suppress warning
            statements guarding against misuse printed to stdout.
        :type suppress_warnings: bool
        """
        api_key, site = get_preferred_credentials(api_key, site)
        self.models = ModelsClient(api_key, site, suppress_warnings=suppress_warnings)
        self.search = SearchClient(api_key, site, suppress_warnings=suppress_warnings)
        self.data = DataClient(api_key, site, suppress_warnings=suppress_warnings)

        clients = [self.models, self.search, self.data]

        for client in clients:
            client_methods = [a for a in dir(client) if not a.startswith('_')]
            for method in client_methods:
                setattr(self, method, _generate_lambda_proxy_method(client, method))


    def __repr__(self):
        return "['models', 'search', 'data']"